import os
import shutil


def post_render_project(configurator):
    variables = configurator.variables
    root = configurator.target_directory
    root = os.path.join(root, variables['project.name'])
    if variables['odoo.series'] in ('8.0', '9.0'):
        shutil.rmtree(os.path.join(root, 'odoo'))
    else:
        shutil.rmtree(os.path.join(root, 'odoo_addons'))



# -*- coding: utf-8 -*-
# Copyright 2016 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import click

from .main import main
from .tools import check_call, call, check_output


def do_tag(config, force, yes):
    tag = config.version
    if not yes:
        click.confirm('Tag project with {}?'.format(tag), abort=True)
    if force:
        force_cmd = ['-f']
    else:
        force_cmd = []
    if 0 != call(['git', 'diff', '--exit-code']):
        raise click.ClickException("Please commit first.")
    if 0 != call(['git', 'diff', '--exit-code', '--cached']):
        raise click.ClickException("Please commit first.")
    out = check_output(['git', 'ls-files', '--other', '--exclude-standard',
                        '--directory'])
    if out:
        click.echo(out)
        raise click.ClickException("Please commit first.")
    check_call(['git', 'tag'] + force_cmd + [tag])
    check_call(['git', 'push'] + force_cmd + ['origin', 'tag', tag])


@click.command(help='Tag the current project after ensuring '
                    'everything has been commited to git.')
@click.option('-f', '--force', is_flag=True,
              help='Replace an existing tag (instead of failing)')
@click.option('-y', '--yes', is_flag=True, default=False)
@click.pass_context
def tag(ctx, force, yes):
    do_tag(ctx.obj['config'], force, yes)


main.add_command(tag)



# -*- coding: utf-8 -*-
# Copyright 2016 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import logging
from pkg_resources import get_distribution, DistributionNotFound

import click

from .config import AcsooConfig

try:
    __version__ = get_distribution('acsoo').version
except DistributionNotFound:
    # package is not installed
    pass

__notice__ = '''%(prog)s, version %(version)s

Acsone Odoo Development Tools.

Copyright 2016-2017 ACSONE SA/NV (<http://acsone.eu>)
License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).'''


@click.group()
@click.version_option(version=__version__, message=__notice__)
@click.option('-v', '--verbose', count=True)
@click.option('-c', '--config', type=click.Path(),
              help="Configuration file (default: ./acsoo.cfg).")
@click.pass_context
def main(ctx, verbose, config):
    config = AcsooConfig(config)

    ctx.obj = dict(config=config)

    ctx.default_map = config.get_default_map()

    if verbose > 1:
        level = logging.DEBUG
    elif verbose > 0:
        level = logging.INFO
    else:
        level = logging.WARNING
    logging.basicConfig(format='%(message)s', level=level)



# -*- coding: utf-8 -*-
# Copyright 2016 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

# flake8: noqa

from . import main
# commands
from . import addons
from . import checklog
from . import flake8cmd
from . import pylintcmd
from . import release
from . import tag
from . import tag_editable_requirements
from . import wheel



# -*- coding: utf-8 -*-
# Copyright 2017 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import os
import unittest

from click.testing import CliRunner

from acsoo.addons import addons
from acsoo.tools import working_directory


DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


class TestAddons(unittest.TestCase):

    def test_list(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(addons, [
                'list',
            ])
            self.assertTrue(res.exit_code == 0)
            expected = 'addon1'
            assert expected in res.output

    def test_exclude(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(addons, [
                '--exclude', 'addon1',
                'list',
            ])
            self.assertTrue(res.exit_code == 0)
            expected = ''
            assert expected in res.output

    def test_include(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(addons, [
                '--include', 'addon1',
                'list',
            ])
            self.assertTrue(res.exit_code == 0)
            expected = 'addon1'
            assert expected in res.output

    def test_list_depends(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(addons, [
                'list-depends',
            ])
            self.assertTrue(res.exit_code == 0)
            expected = 'base'
            assert expected in res.output



# -*- coding: utf-8 -*-
# Copyright 2016 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import logging
import os
import shutil

import click

from .main import main
from .tools import check_call, working_directory

_logger = logging.getLogger(__name__)


def do_wheel(src, requirement, wheel_dir, no_cache_dir, no_index):
    if os.path.exists(wheel_dir):
        _logger.debug('Removing all wheels in %s.', wheel_dir)
        with working_directory(wheel_dir):
            for f in os.listdir('.'):
                if f.endswith('.whl'):
                    os.remove(f)
    opts = []
    if no_cache_dir:
        opts.append('--no-cache-dir')
    if no_index:
        opts.append('--no-index')
    check_call(['pip', 'wheel', '--src', src, '-r', 'requirements.txt',
                '--wheel-dir', wheel_dir] + opts)
    # TODO 'pip wheel .' is slower and sometimes buggy because of
    #      https://github.com/pypa/pip/issues/3499
    if os.path.exists('build'):
        shutil.rmtree('build')
    check_call(['python', 'setup.py', 'bdist_wheel',
                '--dist-dir', wheel_dir] + opts)


@click.command(help='Build wheels for all dependencies found in '
                    'requirements.txt, plus the project in the current '
                    'directory. CAUTION: all wheel files are removed from '
                    'the target directory before building')
@click.option('--src', default='src', envvar='PIP_SRC',
              type=click.Path(),
              help='Directory where editable requirements are checked out')
@click.option('-r', '--requirement', default='requirements.txt',
              type=click.File(),
              help='Requirements to build (default=requirements.txt)')
@click.option('-w', '--wheel-dir', default='release',
              type=click.Path(),
              help='Path where the wheels will be created (default=release')
@click.option('--no-cache-dir', is_flag=True,
              help='Disable the pip cache')
@click.option('--no-index', is_flag=True,
              help='Ignore package index '
                   '(only looking at --find-links URLs instead)')
def wheel(src, requirement, wheel_dir, no_cache_dir, no_index):
    do_wheel(src, requirement, wheel_dir, no_cache_dir, no_index)


main.add_command(wheel)



# -*- coding: utf-8 -*-
# Copyright 2016 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import logging
import re
import os

import click

from .main import main
from .tools import check_call, working_directory

_logger = logging.getLogger(__name__)


RE = re.compile(r"^-e git\+(?P<url>ssh://.*?/.*?)@(?P<sha>[^?#&]+)"
                r".*?[#?]egg=(?P<egg>[^?#&]+)")


def do_tag_editable_requirements(config, force, src, requirement, yes):
    tag = '{}-{}_{}'.format(
        config.series, config.trigram, config.version)
    if not yes:
        click.confirm('Tag dependencies with {}?'.format(tag), abort=True)
    if force:
        force_cmd = ['-f']
    else:
        force_cmd = []
    shas = {}  # url: sha
    for req in requirement:
        req = req.strip()
        mo = RE.match(req)
        if not mo:
            continue
        url = mo.group('url')
        sha = mo.group('sha')
        egg = mo.group('egg')
        if url in shas:
            if shas[url] != sha:
                raise click.ClickException(
                    "Trying to place tag {tag} at {url}@{sha} but the "
                    "same tag has already been placed at {prevsha}. This "
                    "is probably due to an inconsistency in your "
                    "requirements.txt.".format(
                        tag=tag, url=url, sha=sha, prevsha=shas[url]))
            # we already tagged this url with this sha
            _logger.info("Skipping %s@%s because already tagged.", url, sha)
            continue
        repo = os.path.join(src, egg.replace('_', '-'))
        if not os.path.isdir(os.path.join(repo, '.git')):
            check_call(['git', 'clone', url, repo])
        with working_directory(repo):
            _logger.info('placing tag %s on %s@%s', tag, url, sha)
            check_call(['git', 'fetch', '--tags', url, sha])
            check_call(['git', 'tag'] + force_cmd + [tag, sha])
            check_call(['git', 'push'] + force_cmd + [url, tag])
        shas[url] = sha


@click.command(help='Tag all editable requirements found in '
                    'requirements.txt, so the commits referenced in '
                    'there are not lost in case of git garbage collection.')
@click.option('-f', '--force', is_flag=True,
              help='Replace an existing tag (instead of failing)')
@click.option('--src', default='src', envvar='PIP_SRC',
              type=click.Path(),
              help='Directory where editable requirements are checked out')
@click.option('-r', '--requirement', default='requirements.txt',
              type=click.File(),
              help='Requirements to build (default=requirements.txt)')
@click.option('-y', '--yes', is_flag=True, default=False)
@click.pass_context
def tag_editable_requirements(ctx, force, src, requirement, yes):
    do_tag_editable_requirements(
        ctx.obj['config'], force, src, requirement, yes)


main.add_command(tag_editable_requirements)



# -*- coding: utf-8 -*-
# Copyright 2017 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import os

import click

from .main import main
from .manifest import get_installable_addons


@click.group(help="Do things with addons lists. Options of this command "
                  "select addons on which the subcommands will act.")
@click.option('--addons-dir', 'addons_dirs', multiple=True, type=click.Path(),
              help="Directory containing addons. Defaults to odoo/addons or "
                   "odoo_addons if present. This option can be repeated.")
@click.option('--include', default='',
              help="Comma separated list of addons to include (default: all "
                   "installable addons found in --addons-dir').")
@click.option('--exclude', default='',
              help="Comma separated list of addons to exclude.")
@click.pass_context
def addons(ctx, addons_dirs, include, exclude):
    include = set(s.strip() for s in include.split(',') if s.strip())
    exclude = set(s.strip() for s in exclude.split(',') if s.strip())
    if not addons_dirs:
        addons_dirs = []
        candidate_addons_dirs = (
            os.path.join('odoo', 'addons'),
            os.path.join('odoo_addons'),
        )
        for addons_dir in candidate_addons_dirs:
            if os.path.isdir(addons_dir):
                addons_dirs.append(addons_dir)
    manifests = {}
    for addons_dir in addons_dirs:
        for addon, manifest in get_installable_addons(addons_dir).items():
            if (not include or addon in include) and addon not in exclude:
                manifests[addon] = manifest
    ctx.obj = dict(manifests=manifests)


main.add_command(addons)


@click.command(help="Print a comma separated list of selected addons.")
@click.pass_context
def addons_list(ctx):
    manifests = ctx.obj['manifests']
    addon_names = sorted(manifests.keys())
    click.echo(','.join(addon_names))


addons.add_command(addons_list, 'list')


@click.command(help="Print a comma separated list of the direct dependencies "
                    "of installable addons found in --addons-dir.")
@click.pass_context
def addons_list_depends(ctx):
    manifests = ctx.obj['manifests']
    depends = set()
    for manifest in manifests.values():
        depends.update(manifest.get('depends', []))
    depends -= set(manifests.keys())
    addon_names = sorted(depends)
    click.echo(','.join(addon_names))


addons.add_command(addons_list_depends, 'list-depends')



# -*- coding: utf-8 -*-
# Copyright 2017 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import os
import unittest
from textwrap import dedent

from click.testing import CliRunner

from acsoo.main import main
from acsoo.pylintcmd import pylintcmd
from acsoo.tools import working_directory


DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


class TestPylint(unittest.TestCase):

    def test1(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(pylintcmd, [
                '-e', 'fixme:0,manifest-required-key',
                '-m', 'odoo',
            ])
            self.assertTrue(res.exit_code != 0)
            expected = dedent("""\
                messages that did not cause failure:
                  manifest-required-key: 1
                messages that caused failure:
                  fixme: 1 (expected 0)
            """)
            assert expected in res.output

    def test2(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(pylintcmd, [
                '-e', 'fixme:0,manifest-required-key',
            ])
            self.assertTrue(res.exit_code != 0)
            expected = dedent("""\
                messages that did not cause failure:
                  manifest-required-key: 1
                messages that caused failure:
                  fixme: 1 (expected 0)
            """)
            assert expected in res.output

    def test2_config(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(main, [
                '-c', os.path.join(DATA_DIR, 'test_pylint2.cfg'),
                'pylint',
            ])
            self.assertTrue(res.exit_code != 0)
            expected = dedent("""\
                messages that did not cause failure:
                  manifest-required-key: 1
                messages that caused failure:
                  fixme: 1 (expected 0)
            """)
            assert expected in res.output

    def test3(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(pylintcmd, [
                '--', '-d', 'fixme',
            ])
            self.assertTrue(res.exit_code != 0)
            assert 'fixme: ' not in res.output

    def test3_config(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(main, [
                '-c', os.path.join(DATA_DIR, 'test_pylint3.cfg'),
                'pylint',
            ])
            self.assertTrue(res.exit_code != 0)
            assert 'fixme: ' not in res.output



# -*- coding: utf-8 -*-
# Copyright © 2015 ACSONE SA/NV
# License LGPLv3 (http://www.gnu.org/licenses/lgpl-3.0-standalone.html)

import ast
import os

MANIFEST_NAMES = ('__openerp__.py', '__manifest__.py', '__terp__.py')


class NoManifestFound(Exception):
    pass


def get_manifest_path(addon_dir):
    for manifest_name in MANIFEST_NAMES:
        manifest_path = os.path.join(addon_dir, manifest_name)
        if os.path.isfile(manifest_path):
            return manifest_path


def parse_manifest(s):
    return ast.literal_eval(s)


def get_installable_addons(addons_dir):
    res = {}
    for addon_name in os.listdir(addons_dir):
        addon_dir = os.path.join(addons_dir, addon_name)
        manifest_path = get_manifest_path(addon_dir)
        if not manifest_path:
            continue
        with open(manifest_path) as f:
            manifest = parse_manifest(f.read())
        if not manifest.get('installable', True):
            continue
        res[addon_name] = manifest
    return res



# -*- coding: utf-8 -*-
# Copyright 2017 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import logging

import click

from .config import AcsooConfig
from .main import main
from .tools import check_call, cfg_path


def do_flake8(config, flake8_options):
    # flake8 __init__.py
    cmd = [
        'flake8',
        '--config', cfg_path('flake8__init__.cfg'),
    ]
    check_call(cmd, log_level=logging.INFO)
    # flake8
    cmd = [
        'flake8',
        '--config', config,
    ] + list(flake8_options)
    check_call(cmd, log_level=logging.INFO)


@click.command(help='Run flake8')
@click.option('--config', type=click.Path(), default=cfg_path('flake8.cfg'),
              help="Flake8 config file. Default is provided by acsoo.")
@click.argument('flake8-options', nargs=-1)
def flake8(config, flake8_options):
    do_flake8(config, flake8_options)


main.add_command(flake8)


def _read_defaults(config):
    section = 'pylint'
    defaults = dict(
        config=config.get(
            section, 'config', default=cfg_path('flake8.cfg')),
    )
    return dict(flake8=defaults)


AcsooConfig.add_default_map_reader(_read_defaults)



# -*- coding: utf-8 -*-
# Copyright 2016 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import os
from ConfigParser import RawConfigParser, NoOptionError, NoSectionError

import click


DEFAULT_CONFIG_FILE = 'acsoo.cfg'
SECTION = 'acsoo'


def _split_multiline(s):
    return [i.strip() for i in s.splitlines() if i.strip()]


class AcsooConfig(object):

    # list of callables returning dictionaries to update default_map
    default_map_readers = []

    def __init__(self, filename):
        self.__cfg = RawConfigParser()
        if not filename and os.path.isfile(DEFAULT_CONFIG_FILE):
            filename = DEFAULT_CONFIG_FILE
        if filename:
            if not os.path.isfile(filename):
                raise click.ClickException("Configuration file {} not found.".
                                           format(filename))
            self.__cfgfile = filename
            self.__cfg.read(filename)

    @staticmethod
    def add_default_map_reader(reader):
        AcsooConfig.default_map_readers.append(reader)

    def get_default_map(self):
        default_map = {}
        for reader in self.default_map_readers:
            default_map.update(reader(self))
        return default_map

    @property
    def series(self):
        r = self.__cfg.get(SECTION, 'series')
        if not r:
            raise click.ClickException('Missing series in {}.'.
                                       format(self.__cfgfile))
        if r not in ('8.0', '9.0', '10.0'):
            raise click.ClickException('Unsupported series {} in {}.'.
                                       format(r, self.__cfgfile))
        return r

    @property
    def version(self):
        r = self.__cfg.get(SECTION, 'version')
        if not r:
            raise click.ClickException('Missing version in {}.'.
                                       format(self.__cfgfile))
        return r

    @property
    def trigram(self):
        r = self.__cfg.get(SECTION, 'trigram')
        if not r:
            raise click.ClickException('Missing trigram in {}.'.
                                       format(self.__cfgfile))
        return r

    def get(self, section, option, default=None, flatten=False):
        try:
            r = self.__cfg.get(section, option)
            if flatten:
                r = ''.join(_split_multiline(r))
            return r
        except (NoOptionError, NoSectionError):
            return default

    def getboolean(self, section, option, default=None):
        try:
            return self.__cfg.getboolean(section, option)
        except (NoOptionError, NoSectionError):
            return default

    def getlist(self, section, option, default=None):
        try:
            return _split_multiline(self.__cfg.get(section, option))
        except (NoOptionError, NoSectionError):
            return default or []



# -*- coding: utf-8 -*-
# Copyright 2016-2017 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import contextlib
import logging
import os
import subprocess
import sys
from distutils.spawn import find_executable as _fe

import click

_logger = logging.getLogger(__name__)


def _escape(s):
    s = s.replace('\\', '\\\\')
    s = s.replace('"', '\\"')
    s = s.replace('\'', '\\\'')
    s = s.replace('&', '\\&')
    s = s.replace('|', '\\|')
    s = s.replace('>', '\\>')
    s = s.replace('<', '\\<')
    s = s.replace(' ', '\\ ')
    return s


def cmd_string(cmd):
    return " ".join([_escape(s) for s in cmd])


def log_cmd(cmd, cwd=None, level=logging.DEBUG, echo=False):
    s = cmd_string(cmd)
    if echo:
        click.echo(click.style(s, bold=True))
    _logger.log(level, '%s$ %s', cwd or '.', s)


def call(cmd, cwd=None, log_level=logging.DEBUG, echo=False):
    _adapt_executable(cmd)
    log_cmd(cmd, cwd=cwd, level=log_level, echo=echo)
    try:
        return subprocess.call(cmd, cwd=cwd)
    except subprocess.CalledProcessError:
        raise click.ClickException(cmd_string(cmd))


def check_call(cmd, cwd=None, log_level=logging.DEBUG, echo=False):
    _adapt_executable(cmd)
    log_cmd(cmd, cwd=cwd, level=log_level, echo=echo)
    try:
        return subprocess.check_call(cmd, cwd=cwd)
    except subprocess.CalledProcessError:
        raise click.ClickException(cmd_string(cmd))


def check_output(cmd, cwd=None, log_level=logging.DEBUG, echo=False):
    _adapt_executable(cmd)
    log_cmd(cmd, cwd=cwd, level=log_level, echo=echo)
    try:
        return subprocess.check_output(cmd, cwd=cwd)
    except subprocess.CalledProcessError:
        raise click.ClickException(cmd_string(cmd))


@contextlib.contextmanager
def working_directory(path):
    """A context manager which changes the working directory to the given
    path, and then changes it back to its previous value on exit.
    """
    prev_cwd = os.getcwd()
    _logger.debug('.$ cd %s', _escape(path))
    os.chdir(path)
    try:
        yield
    finally:
        _logger.debug('.$ cd %s', _escape(prev_cwd))
        os.chdir(prev_cwd)


def _adapt_executable(cmd):
    cmd[0] = _find_executable(cmd[0])


def _find_executable(exe):
    python_dir = os.path.dirname(sys.executable)
    exe_path = os.path.join(python_dir, exe)
    if os.path.exists(exe_path):
        return exe_path
    exe_path = _fe(exe)
    if exe_path:
        return exe_path
    raise RuntimeError("%s executable not found" % (exe, ))


def cfg_path(filename):
    return os.path.join(os.path.dirname(__file__), 'cfg', filename)
